# Open a console
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import threading
import time
import libxml2

from .log import log
from .VMDefaults import VMDefaults
from . import msg
from . import util
from . import caps


################################################################################
# GLOBAL DECLARAIONS
################################################################################
_reset_console = False          # 
console        = None           # current console instance


################################################################################
# CLASS: Console
#
# Class to abstract the console.
################################################################################
class Console:
    def __init__(self):
        #console.console = self
        pass

    def parse_xml_data(self, xml, element, sub):
        try:
            doc = libxml2.parseDoc(xml)
        except:
            return None
        try:
            ctxt = doc.xpathNewContext()
            result = ctxt.xpathEval("//"+element)
            if len(result):
                value=str(result[0])
                if sub:
                    data = value.split(sub)
                    quote = None
                    value = ""
                    if data[1]:
                        for d in data[1]:
                            if d == '=':
                                continue
                            if d == '\"' or d == '\'':
                                if quote or d == '>':
                                    break
                            if d == '>':
                                break
                            if d == '\"' or d == '\'':
                                quote = 1
                                continue
                            value = value + d
                return value
            return None 
        finally:
            doc.freeDoc()


    #---------------------------------------------------------------------------
    # Class:  Console
    # Method: get_domid_from_uuid
    # Parms:
    #   uuid:   uuid from the domain
    #
    # This method gets the domain ID for the given UUID.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def get_domid_from_uuid(self, uuid):
        log.debug("Domain %s is not running" % uuid)
        return None

    def wait_for_uuid_to_exit(self, uuid):
        self.wait_for_domid_to_exit(self.get_domid_from_uuid(uuid), uuid)


    #---------------------------------------------------------------------------
    # Class:  Console
    # Method: wait_for_domid_to_exit
    #
    # Parms:
    #   domid:  The domain ID of the domain that we are waiting to exit.
    #   uuid:   UUID from the domain
    #
    # This method waits for the given domain to exit.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def wait_for_domid_to_exit(self, domid, uuid):
        """Wait for domid to exit (or, for domid's UUID to not match
        the specified uuid, which would indicate a race condition in
        which domid has already exited and was replace by another)."""
        return


    #---------------------------------------------------------------------------
    # Class:  Console
    # Method: get_vnc_port_from_domid
    #
    # Parms:
    #   domid:      The domain ID of the domain that we are waiting to exit.
    #   timeout:    Wait only so long.
    #
    # This method finds the vnc port for the given domain.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def get_vnc_port_from_domid(self, domid, timeout=0.0):
        return None


    #---------------------------------------------------------------------------
    # Class:  Console
    # Method: get_tty_from_domid
    # Parms:
    #   domid:  domain ID
    #
    # This method gets the tty number for a given domain.
    #
    # This method is stubbed out in the base class.  It is expected that
    # anyone that subclasses this class will create their own method.
    #---------------------------------------------------------------------------
    def get_tty_from_domid(self, domid):
        return None


    def vnc_console(self, guest, detach=False):
        """Forks a vncviewer process to view the domain.  Returns the
        child's PID if (success and not detach), else None.  Note that
        failure cannot be determined if detach is True.
        """
        if guest.options.graphics_viewer == 'vnc':
            default_viewer = 'vncviewer'
        else:
            default_viewer = 'virt-viewer'
        domid = self.get_domid_from_uuid(guest.uuid)
        vncport = self.get_vnc_port_from_domid(domid, timeout=10.0)
        if vncport == None:
            log.info("Did not find VNC port, so not using VNC.")
            raise NotImplementedError
        if 'DISPLAY' not in os.environ:
            raise RuntimeError(msg.vnc_no_display % ('localhost', vncport))
        def find_viewer():
            try:
                for path in os.environ['PATH'].split(':'):
                    viewer = os.path.join(path, default_viewer)
                    if os.path.exists(viewer):
                        log.info("Found %s: %s" % (default_viewer, viewer))
                        return viewer
            except:
                pass
            return None
        def exec_viewer(viewer):
            try:
                if default_viewer == 'virt-viewer':
                    os.execv(viewer, [default_viewer, '%d' % domid, ])
                else:
                    os.execv(viewer, [default_viewer, '::%d' % vncport, ])
            except:
                os._exit(1)
        viewer = find_viewer()
        if not viewer:
            raise RuntimeError(msg.display_failed)
        if not detach:
            pid = os.fork()
            if pid == 0:
                # Closing all fds prevents virt-viewer from launching
                if default_viewer == 'virt-viewer':
                    fds=range(1,256)
                else:
                    fds=range(0,256)
                util.close_fds(fds)
                exec_viewer(viewer)
        else:
            if default_viewer == 'virt-viewer':
                fds=range(1,256)
            else:
                fds=range(0,256)
            if util.daemonize(fds):
                time.sleep(1)
                exec_viewer(viewer)
            pid = None
        log.info("Opened a VNC viewer to %s:%d." % ('localhost', vncport))
        return pid

    def text_console(self, guest, detach=False):
        """Forks a text console process to view the domain.  Returns the
        child's PID if (success and not detach), else None.  Note that
        failure cannot be determined if detach is True.
        """
        domid = self.get_domid_from_uuid(guest.uuid)
        tty = self.get_tty_from_domid(domid)
        if tty is None:
            raise NotImplementedError
        if not os.access(tty, os.R_OK | os.W_OK):
            log.debug("%s may not be accessible, attempting to use anyways" % tty)
        if 'DISPLAY' in os.environ:
            vmdefaults = VMDefaults[guest.options.os_type]
            cmd = '/usr/bin/xterm'
            args = ['/usr/bin/xterm']
            args.extend(vmdefaults.xterm_args())
            if caps.is_kvm() or caps.is_qemu():
                args.extend(['-T', guest.options.name, '-e', 'virsh', 'console', guest.uuid])
            else:
                if util.is_xend_running():
                    args.extend(['-T', guest.options.name, '-e', 'xm', 'console', guest.uuid])
                else:
                    args.extend(['-T', guest.options.name, '-e', 'xl', 'console', str(domid)])
            log.info("Opened a console in an xterm.")
        else:
            if caps.is_kvm() or caps.is_qemu():
                cmd = '/usr/bin/virsh'
                args = ['/usr/sbin/virsh', 'console', guest.uuid ]
            else:
                if util.is_xend_running():
                    cmd = '/usr/sbin/xm'
                    args = ['/usr/sbin/xm', 'console', guest.uuid ]
                else:
                    cmd = '/usr/sbin/xl'
                    args = ['/usr/sbin/xl', 'console', str(domid)]
        log.debug("Running: %s" % str(args))
        if not detach:
            pid = os.fork()
            if pid == 0:
                os.execvp(cmd, args)
                os._exit(1)
        else:
            if util.daemonize():
                os.execvp(cmd, args)
                os._exit(1)
            pid = None
        return pid

def reset():
    global _reset_console
    if _reset_console:
        os.system('reset')
        _reset_console = False
