# Direct Kernel Boot guest support
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import re

from .VMDefaults import VMDefaults
from .exceptions import *
from .globals import *
from .log import log
from . import mount
from . import paths
from . import util
import vmdisks


class DirectKernelBoot:
    def __init__(self, options):
        self.kernel = ''
        self.initrd = ''
        self.options = options

    def _get_partitions(self, dev):
        """Returns a list of dictionaries that describe the (relevant)
        partitions on the device.
        """
        found = False
        partitions = []
        pre = re.compile('([^ ]*)[ \\t]*:[ \\t]*start=[ \\t]*([0-9]*)[ \\t]*, size=[ \\t]*([0-9]*)[ \\t]*, Id=[ \\t]*([0-9a-fA-F]*)')
        log.debug("Examining '%s' for partitions..." % dev)
        lines = util.run(['/sbin/sfdisk', '--force', '-d', '-uS', '--', dev])
        for line in lines.splitlines():
            m = pre.match(line)
            if m:
                partition = {}
                pname = m.group(1)
                pstart = m.group(2)
                pid = m.group(4)
                # For now, only investigate 83 (Linux) and 6 (FAT16)
                if pid not in ('6', '83'):
                    continue
                log.debug("Found partition '%s'." % (pname,))
                partition['name'] = pname
                partition['id'] = pid
                partition['offset'] = int(pstart) * 512
                partitions.append(partition)
                found = True

        # Try '8e' (Linux LVM) scan
        if not found:
            log.debug("Try LVM scan: vdev %s" % (dev,))
            vgname = []
            fd = os.popen("pvscan | awk '$2 == \"%s\" {print $4}' 2>/dev/null" % (dev))
            for line in fd.readlines():
                line = line.strip()
                vgname = line[:]
            fd.close()
            if not vgname:
                return partitions

            log.debug("LVM Found vg: vdev %s" % (vgname))
            ret = os.system("vgchange -ay '%s' > /dev/null 2>&1" % (vgname)) >> 8
            if not ret:
                log.debug("lvscan | grep '/dev/%s'" % (vgname))
                fd = os.popen("lvscan | grep '/dev/%s'" % (vgname))
                for line in fd.readlines():
                    line = line.strip()
                    (t1, lvname, t2) = line.split('\'')
                    pname = lvname[lvname.rfind('/')+1:]
                    pname = pname.strip()
                    pname = "/dev/mapper/" + vgname + "-" + pname
                    partition = {}
                    partition['name'] = pname
                    partition['id'] = '8e'
                    partition['offset'] = 0
                    partition['vdev'] = vgname
                    partitions.append(partition)
                    log.debug("Found partition: vdev %s, pdev %s" % (vgname, pname))
                fd.close()
                log.debug("vgchange -an '%s'" % (vgname))
                os.system("vgchange -an '%s' > /dev/null 2>&1" % (vgname))
        return partitions

    def _get_directkernelboot_images(self, dev, partitions, copy, progress=lambda f:None):
        def traildigits(strg):
            """Return the trailing digits, used to split the partition number off"""
            idx = len(strg)-1
            while strg[idx].isdigit():
                if len == 0:
                    return strg
                idx -= 1
            return strg[idx+1:]
        actions = VMDefaults[self.options.os_type].actions()
        for partition in partitions:
            # For now, only investigate 6 (FAT16) and 83 (Linux), 8e (Linux LVM)
            if partition['id'] not in ('6', '83', '8e'):
                continue
            if partition['id'] == '8e':
                os.system("vgchange -ay '%s' > /dev/null 2>&1" % (partition['vdev']))
                dev = partition['name']
            log.debug("Loop-mounting '%s', offset %d..." % (dev, partition['offset']))
            vdisk = mount.TemporaryLoopMount(dev)
            try:
                mountpoint = vdisk.mount(options='loop,ro,offset=%d' % partition['offset'])
            except Exception as e:
                log.debug("Mount failed (%s); skipping." % str(e))
                if partition['id'] == '8e':
                    os.system("vgchange -an '%s' > /dev/null 2>&1" % (partition['vdev']))
                continue
            try:
                (_kernel, _initrd) = actions.getRuntimeImagesFromPath(mountpoint, self.options.os_type)
                if _kernel:
                    if partition['id'] == '8e':
                        vpart = 'LVM-' + partition['name'][partition['name'].rfind('/')+1:]
                    else:
                        vpart = traildigits(partition['name'])
                    log.info("Found kernel on partition %s." % vpart)
                    if copy:
                        # Be careful to explicitly close on failure, and close
                        # immediately after use, otherwise no guarantee these
                        # will be closed by GC before attempting unmount.
                        kfd = open(mountpoint + _kernel, 'rb')
                        try:
                            ifd = None
                            if _initrd:
                                ifd = open(mountpoint + _initrd, 'rb')
                        except:
                            kfd.close()
                            raise
                        try:
                            kernel = util.copy_to_temp(kfd, prefix='kernel.')
                            if not ifd:
                                initrd = None
                            else:
                                try:
                                    initrd = util.copy_to_temp(ifd, prefix='initrd.')
                                except:
                                    try: os.unlink(kernel)
                                    except: pass
                                    raise
                        finally:
                            kfd.close()
                            if ifd:
                                ifd.close()
                    else:
                        log.info("Not copying; will use kernel/initrd in-place")
                        kernel = _kernel
                        initrd = _initrd
                    log.info("kernel %s; initrd %s" % (kernel, initrd))
                    self.kernel = kernel
                    self.initrd = initrd
                    return (vpart, kernel, initrd)
            finally:
                vdisk.unmount()
                if partition['id'] == '8e':
                    log.debug("vgchange -an '%s'" % (partition['vdev']))
                    os.system("vgchange -an '%s' > /dev/null 2>&1" % (partition['vdev']))
                progress(1.0)
        progress(1.0)
        return (None, None, None)

    def _get_directkernelboot_install_images(self, progress=lambda f:None):
        """Retrieves the kernel and initrd from the installation source.
        Potentially a heavy-weight operation."""
        actions = VMDefaults[self.options.os_type].actions()
        source = self.options.source
        nonpae = VMDefaults[self.options.os_type].supports_32nonpae()
        if type(source) is vmdisks.Disk:
            log.debug("Installation source: %s" % source)
            if not source.exists():
                raise InstSrcError(InstSrcError.E_NOT_FOUND, details=str(source))
            vdev = source.block_attach()
            try:
                partitions = self._get_partitions(vdev)
                if len(partitions) == 0:
                    mounted_disk = mount.TemporaryLoopMount(vdev)
                    try:
                        mountpoint = mounted_disk.mount()
                    except Exception as e:
                        raise InstSrcError(InstSrcError.E_BAD, details=e.args[0])
                    try:
                        (kfn, ifn) = actions.getInstallImagesFromPath(self.options, mountpoint, progress, nonpae)
                    finally:
                        mounted_disk.unmount()
                else:
                    (_, kfn, ifn) = self._get_directkernelboot_images(vdev, partitions, copy=True, progress=progress)
            finally:
                source.block_detach()
        else:
            (kfn, ifn) = actions.getInstallImagesFromUrl(self.options, source, progress, nonpae)
        if kfn is None:
            raise InstSrcError(InstSrcError.E_BAD, details=str(source))
        self.kernel = kfn
        self.initrd = ifn
        return (kfn, ifn)

    def _cleanup_kernel(self):
        if self.options.preserve_on_error:
            log.info("Installation failed, not cleaning up disks because preserve_on_error is set")
            self.kernel = ''
            self.initrd = ''
            return
        if self.kernel:
            try:
                os.unlink(self.kernel)
            finally:
                self.kernel = ''
        if self.initrd:
            try:
                os.unlink(self.initrd)
            finally:
                self.initrd = ''

